import {
  ComponentFactoryResolver,
  ComponentRef,
  Directive,
  ElementRef,
  EventEmitter,
  Input,
  NgZone,
  OnChanges,
  OnDestroy,
  Output,
  Renderer2,
  SimpleChanges,
  TemplateRef,
  ViewContainerRef,
  forwardRef,
} from "@angular/core";
import {
  AbstractControl,
  ControlValueAccessor,
  NG_VALIDATORS,
  NG_VALUE_ACCESSOR,
  Validator,
} from "@angular/forms";
import { EventManager } from "@isp/xdce-arch-core-base";
import {
  DayTemplateContextNOA11Y,
  NbpCalendarPattern,
  NbpCalendarPosition,
  NbpDateSeparator,
  NgbCalendarNOA11Y,
  NgbDate,
  NgbDateParserFormatterNOA11Y,
  NgbDateStruct,
  NgbDatepickerNavigateEventNOA11Y,
  NgbDatepickerServiceNOA11Y,
} from "@isp/xdce-widget";
import { Positioning } from "../util/positioning";
import {
  FidRplCalendarNOA11Y,
} from "./datepicker";

const NGB_DATEPICKER_VALUE_ACCESSOR = {
  provide: NG_VALUE_ACCESSOR,
  useExisting: forwardRef(() => NgbFidInputDatepickerNOA11Y),
  multi: true,
};

const NGB_DATEPICKER_VALIDATOR = {
  provide: NG_VALIDATORS,
  useExisting: forwardRef(() => NgbFidInputDatepickerNOA11Y),
  multi: true,
};

/**
 * A directive that makes it possible to have datepickers on input fields.
 * Manages integration with the input field itself (data entry) and ngModel (validation etc.).
 */
@Directive({
  selector: "input[fidRplCalendarNOA11Y]",
  exportAs: "fidRplCalendarNOA11Y",
  host: {
    "(input)": "waitForFormat($event.target.value)",
    "(change)": "manualDateChange($event.target.value, true)",
    "(keyup.esc)": "close()",
    "(blur)": "onBlur()",
    "(document:click)": "onFocusOut($event)",
  },
  providers: [
    NGB_DATEPICKER_VALUE_ACCESSOR,
    NGB_DATEPICKER_VALIDATOR,
    NgbDatepickerServiceNOA11Y,
  ],
})
export class NgbFidInputDatepickerNOA11Y
  implements OnChanges, OnDestroy, ControlValueAccessor, Validator
{
  private _cRef: ComponentRef<FidRplCalendarNOA11Y> = null;
  private _model: NgbDate;
  private _zoneSubscription: any;

  opened: boolean;

  dataBeforeChange;
  /**
   * Reference for the custom template for the day display
   */
  @Input() dayTemplate: TemplateRef<DayTemplateContextNOA11Y>;

  /**
   * Number of months to display
   */
  @Input() displayMonths: number;

  /**
   * First day of the week. With default calendar we use ISO 8601: 1=Mon ... 7=Sun
   */
  @Input() firstDayOfWeek: number;

  /**
   * Callback to mark a given date as disabled.
   * 'Current' contains the month that will be displayed in the view
   */
  @Input() markDisabled: (
    date: NgbDateStruct,
    current: { year: number; month: number }
  ) => boolean;

  /**
   * Min date for the navigation. If not provided will be 10 years before today or `startDate`
   */
  _minDate: NgbDateStruct;
  @Input()
  get minDate() {
    return this._minDate;
  }
  set minDate(minDate: NgbDateStruct) {
    this._minDate = minDate;
  }

  /**
   * Max date for the navigation. If not provided will be 10 years from today or `startDate`
   */
  _maxDate: NgbDateStruct;
  @Input()
  get maxDate() {
    return this._maxDate;
  }
  set maxDate(maxDate: NgbDateStruct) {
    this._maxDate = maxDate;
  }

  /**
   * Navigation type: `select` (default with select boxes for month and year), `arrows`
   * (without select boxes, only navigation arrows) or `none` (no navigation at all)
   */
  @Input() navigation: "select" | "arrows" | "none";

  /**
   * The way to display days that don't belong to current month: `visible` (default),
   * `hidden` (not displayed) or `collapsed` (not displayed with empty space collapsed)
   */
  @Input() outsideDays: "visible" | "collapsed" | "hidden";

  /**
   * Placement of a datepicker popup. Accepts: "top", "bottom", "left", "right", "bottom-left",
   * "bottom-right" etc.
   */
  @Input() placement = "bottom-left";

  /**
   * Whether to display days of the week
   */
  @Input() showWeekdays: boolean;

  /**
   * Whether to display week numbers
   */
  @Input() showWeekNumbers: boolean;

  /**
   * Date to open calendar with.
   * With default calendar we use ISO 8601: 'month' is 1=Jan ... 12=Dec.
   * If nothing or invalid date provided, calendar will open with current month.
   * Use 'navigateTo(date)' as an alternative
   */
  @Input() startDate: { year: number; month: number };

  @Input() disabledWeekend: boolean;

  @Input() holidays: Array<NgbDateStruct>;

  @Input() taxDeadlines: Array<NgbDateStruct>;

  /**
   *  input for unifiedVg
   */
  @Input() isUnifiedVg: boolean;

  private _pattern: any = NbpCalendarPattern;
  @Input() nbpLabelPattern: NbpCalendarPattern = this._pattern.GGMMAAAA;

  private _separator = NbpDateSeparator;
  @Input() nbpDateSeparator: NbpDateSeparator = this._separator.DOT;

  /**
   * An event fired when navigation happens and currently displayed month changes.
   * See NgbDatepickerNavigateEvent for the payload info.
   */
  @Output() navigate = new EventEmitter<NgbDatepickerNavigateEventNOA11Y>();

  private _onChange = (_: any) => {};
  private _onTouched = () => {};
  private _validatorChange = () => {};

  private _placement: any = NbpCalendarPosition;
  private positionService: Positioning;

  private get isInModal(): boolean {
    let el: HTMLElement = <HTMLElement>this._elRef.nativeElement;

    if (!el || !el.parentElement || !el.parentElement.tagName) {
      return false;
    }
    let tagName: string = el.parentElement.tagName;

    while (!tagName.match(/modal/i) && !tagName.match(/body/i)) {
      el = el.parentElement;
      tagName =
        el.parentElement && el.parentElement.tagName
          ? el.parentElement.tagName
          : "";
    }
    return tagName && !!tagName.match(/modal/i);
  }

  private get isCorporate(): boolean {
    return this._cRef.location.nativeElement.classList.contains(
      "isp-corporate-mode-wrapper"
    );
  }

  get _isNative(): boolean {
    return /(ip(a|o)d|iphone|android)/gi.test(window.navigator.userAgent);
  }

  constructor(
    private _parserFormatter: NgbDateParserFormatterNOA11Y,
    private _elRef: ElementRef,
    private _vcRef: ViewContainerRef,
    private _renderer: Renderer2,
    private _cfr: ComponentFactoryResolver,
    ngZone: NgZone,
    private _service: NgbDatepickerServiceNOA11Y,
    private _calendar: NgbCalendarNOA11Y,
    private eventManager: EventManager
  ) {
    this.positionService = new Positioning();
    this._zoneSubscription = ngZone.onStable.subscribe(() =>
      this.setDatepickerPosition()
    );
  }

  setDatepickerPosition() {
    if (!this._cRef || !this._elRef) {
      return;
    }
    const calendar: HTMLElement = this._cRef.location.nativeElement;
    const datepickerInput: HTMLElement = this._elRef.nativeElement;
    const calendarClone: HTMLElement = <HTMLElement>calendar.cloneNode(true);

    this._renderer.removeClass(
      calendarClone.querySelector(".isp-main-datepicker"),
      "d-none"
    );
    this._renderer.addClass(calendarClone, "invisible");
    window.document.body.appendChild(calendarClone);
    const navigation: number = calendarClone.querySelector(
      ".isp-datepicker-navigation"
    )["offsetHeight"];
    const routingView: number = calendarClone.querySelector(
      ".isp-datepicker-routing-view"
    )["offsetHeight"];
    const position: ClientRect = this.positionService.positionElements(
      datepickerInput,
      calendarClone,
      this.placement
    );

    this._renderer.removeChild(calendarClone.parentElement, calendarClone);
    switch (this.placement) {
      case this._placement.TOP:
      case this._placement.TOP_RIGHT:
      case this._placement.TOP_LEFT:
        position.top = this.isCorporate
          ? -navigation - routingView
          : -navigation - routingView - 3;
        break;
      case this._placement.RIGHT:
      case this._placement.LEFT:
        position.top = -navigation - routingView / 2;
        break;
    }
    calendar.style.top = `${position.top}px`;
    calendar.style.left = `${position.left}px`;
    this._renderer.removeClass(
      calendar.querySelector(".isp-main-datepicker"),
      "d-none"
    );
  }

  registerOnChange(fn: (value: any) => any): void {
    this._onChange = fn;
  }

  registerOnTouched(fn: () => any): void {
    this._onTouched = fn;
  }

  registerOnValidatorChange(fn: () => void): void {
    this._validatorChange = fn;
  }

  setDisabledState(isDisabled: boolean): void {
    this._renderer.setProperty(
      this._elRef.nativeElement,
      "disabled",
      isDisabled
    );
    if (this.isOpen()) {
      this._cRef.instance.setDisabledState(isDisabled);
    }
  }

  validate(c: AbstractControl): { [key: string]: any } {
    const value = c.value;

    if (value === null || value === undefined) {
      return null;
    }

    if (typeof value === "string") {
      let newValue: NgbDate;
      newValue = new NgbDate(
        parseInt(value.substring(0, 4), 10),
        parseInt(value.substring(5, 7), 10),
        parseInt(value.substring(8, 10), 10)
      );
      if (!this._calendar.isValid(newValue)) {
        return { ngbDate: { invalid: c.value } };
      }

      if (
        this.minDate &&
        NgbDate.from(newValue).before(NgbDate.from(this.minDate))
      ) {
        return { ngbDate: { requiredBefore: this.minDate } };
      }

      if (
        this.maxDate &&
        NgbDate.from(newValue).after(NgbDate.from(this.maxDate))
      ) {
        return { ngbDate: { requiredAfter: this.maxDate } };
      }
    } else {
      if (!this._calendar.isValid(value)) {
        return { ngbDate: { invalid: c.value } };
      }

      if (
        this.minDate &&
        NgbDate.from(value).before(NgbDate.from(this.minDate))
      ) {
        return { ngbDate: { requiredBefore: this.minDate } };
      }

      if (
        this.maxDate &&
        NgbDate.from(value).after(NgbDate.from(this.maxDate))
      ) {
        return { ngbDate: { requiredAfter: this.maxDate } };
      }
    }
  }

  writeValue(value: any) {
    const ngbDate = value
      ? new NgbDate(value.year, value.month, value.day)
      : null;
    this._model = this._calendar.isValid(value) ? ngbDate : null;
    this._writeModelValue(this._model);
  }

  waitForFormat(value: string, updateView = false) {
    this.dataBeforeChange = value;
    let actualValue = value;
    setTimeout(() => {
      if (actualValue === this.dataBeforeChange) {
        this.manualDateChange(actualValue, updateView);
      }
    }, 2000);
  }

  manualDateChange(value: string, updateView = false) {
    let datesplitted = value.split(/[./-]/g);
    if (
      this.nbpLabelPattern == this._pattern.MMAAAA &&
      datesplitted.length < 3 &&
      datesplitted.length > 1
    ) {
      value = "01." + value;
    }
    let [year, month, day] = value.split("-");
    if (this.isUnifiedVg && this._isNative) {
      if (year && month && day) {
        // value = day + '.' + month + '.' + year;
      }
    }
    this._model = this._service.toValidDate(
      this._parserFormatter.parse(value, this.nbpLabelPattern),
      null
    );
    this._onChange(
      this._model ? this._model.toStruct() : value === "" ? null : value
    );
    if (updateView && this._model) {
      this._writeModelValue(this._model);
    }
  }

  isOpen() {
    return !!this._cRef;
  }

  /**
   * Opens the datepicker with the selected date indicated by the ngModel value.
   */
  open(mode: boolean) {
    if (!this.isOpen()) {
      if (this._model && (this.minDate || this.maxDate)) {
        if (
          (this.minDate &&
            NgbDate.from(this._model).before(NgbDate.from(this.minDate))) ||
          (this.maxDate &&
            NgbDate.from(this._model).after(NgbDate.from(this.maxDate)))
        ) {
          this._model = null;
        }
      }

      const cf = this._cfr.resolveComponentFactory(FidRplCalendarNOA11Y);
      this._cRef = this._vcRef.createComponent(cf);

      this._applyPopupStyling(this._cRef.location.nativeElement);
      this._cRef.instance.writeValue(this._model);
      this._applyDatepickerInputs(this._cRef.instance);
      this._subscribeForDatepickerOutputs(this._cRef.instance);

      this._cRef.instance.disabledWeekend = this.disabledWeekend;
      this._cRef.instance.holidays = this.holidays;
      this._cRef.instance.taxDeadlines = this.taxDeadlines;

      if (mode) {
        this._renderer.addClass(
          this._cRef.location.nativeElement,
          "isp-corporate-mode-wrapper"
        );
      }
      this._cRef.instance.ngOnInit();

      // date selection event handling
      this._cRef.instance.registerOnChange((selectedDate) => {
        this.writeValue(selectedDate);
        this._onChange(selectedDate);
        if (!this.opened) {
          this.close();
        }
      });

      // focus handling
      this._cRef.instance.focus();
    }
  }

  /**
   * Closes the datepicker popup.
   */
  close() {
    if (this.isOpen()) {
      this._vcRef.remove(this._vcRef.indexOf(this._cRef.hostView));
      this._cRef = null;
      this.eventManager.emit("DATEPICKER_CLOSE", true);
    }
  }

  /**
   * Toggles the datepicker popup (opens when closed and closes when opened).
   */
  toggle(mode: boolean) {
    if (this.isOpen()) {
      this.close();
    } else {
      this.open(mode);
    }
  }

  addMargin(value: string) {
    if (this._cRef) {
      this._renderer.setStyle(
        this._cRef.location.nativeElement,
        "margin-top",
        value
      );
    }
  }

  onFocusOut($event) {
    const parent: HTMLElement = this._elRef.nativeElement.parentElement;
    const icon: HTMLElement = parent.querySelector("i");

    if (
      this.isOpen() &&
      !(
        event.target === icon ||
        this._elRef.nativeElement.contains($event.target) ||
        this._cRef.location.nativeElement.contains($event.target)
      )
    ) {
      this._cRef.instance.focusOutComponent($event);
      console.info("onfocusout close");
      this.close();
      if (this.eventManager) {
        this.eventManager.emit("HANDLER_CALENDAR", {});
      }
    }
  }

  /**
   * Navigates current view to provided date.
   * With default calendar we use ISO 8601: 'month' is 1=Jan ... 12=Dec.
   * If nothing or invalid date provided calendar will open current month.
   * Use 'startDate' input as an alternative
   */
  navigateTo(date?: { year: number; month: number }) {
    if (this.isOpen()) {
      this._cRef.instance.navigateTo(date);
    }
  }

  onBlur() {
    this._onTouched();
  }

  ngOnChanges(changes: SimpleChanges) {
    if (changes["minDate"] || changes["maxDate"]) {
      this._validatorChange();
    }
  }

  ngOnDestroy() {
    this.close();
    this._zoneSubscription.unsubscribe();
  }

  getCalendarHeight() {
    if (this._cRef && this._cRef.instance.datepickerRoutingView) {
      return this._cRef.instance.datepickerRoutingView.getRoutingViewHeight();
    }
    return 0;
  }

  getCalendarWidth() {
    if (this._cRef && this._cRef.instance.datepickerRoutingView) {
      return this._cRef.instance.datepickerRoutingView.getRoutingViewWidth();
    }
    return 0;
  }

  getInputHeight() {
    if (this._elRef) {
      return this._elRef.nativeElement.offsetHeight;
    }
    return 0;
  }

  private _applyDatepickerInputs(datepickerInstance: FidRplCalendarNOA11Y): void {
    [
      "dayTemplate",
      "displayMonths",
      "firstDayOfWeek",
      "markDisabled",
      "minDate",
      "maxDate",
      "navigation",
      "outsideDays",
      "showNavigation",
      "showWeekdays",
      "showWeekNumbers",
    ].forEach((optionName: string) => {
      if ((this as any)[optionName] !== undefined) {
        (datepickerInstance as any)[optionName] = (this as any)[optionName];
      }
    });
    datepickerInstance.startDate = this.startDate || this._model;
  }

  private _applyPopupStyling(nativeElement: any) {
    this._renderer.addClass(nativeElement, "dropdown-menu");
    this._renderer.setStyle(nativeElement, "padding", "0");
  }

  private _subscribeForDatepickerOutputs(
    datepickerInstance: FidRplCalendarNOA11Y
  ) {
    datepickerInstance.navigate.subscribe((date: any) =>
      this.navigate.emit(date)
    );
  }

  private _writeModelValue(model: NgbDate) {
    if (model) {
      const { day, month, year } = model;

      if (this.nbpLabelPattern == this._pattern.MMAAAA) {
        this._renderer.setProperty(
          this._elRef.nativeElement,
          "value",
          this._parserFormatter.format(
            { year, month, day: 0 },
            this.nbpLabelPattern,
            this.nbpDateSeparator
          )
        );
      } else {
        this._renderer.setProperty(
          this._elRef.nativeElement,
          "value",
          this._parserFormatter.format(
            { year, month, day },
            this.nbpLabelPattern,
            this.nbpDateSeparator
          )
        );
      }
    } else {
      this._renderer.setProperty(
        this._elRef.nativeElement,
        "value",
        this._parserFormatter.format(
          model,
          this.nbpLabelPattern,
          this.nbpDateSeparator
        )
      );
    }

    if (this.isOpen()) {
      this._cRef.instance.writeValue(model);
      this._onTouched();
    }
  }
}
